# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""
Module for encapsulating the history of an object tracked over time, estimating its
current state even in case of missing updates and managing its lifecycle
"""

from __future__ import annotations

import logging
from copy import deepcopy
from dataclasses import dataclass
from datetime import datetime, timedelta
from enum import Enum
from typing import Any, Dict, List, Optional, Union

import numpy as np
from filterpy.kalman import KalmanFilter
from mlcvzoo_base.api.data.bounding_box import BoundingBox
from mlcvzoo_base.api.data.box import compute_iou, euclidean_distance
from mlcvzoo_base.api.data.class_identifier import ClassIdentifier

from mlcvzoo_tracker.configuration import (
    KalmanFilterConfig,
    TrackerConfig,
    TrackingToolTrackerConfigObjectSpeed,
)
from mlcvzoo_tracker.types import ImageType

logger = logging.getLogger(__name__)

image_type = Union[np.ndarray]


class TrackerState(Enum):
    """
    A Track has three states:

        1. INITIATED: When an ImageTrack is instantiated
        2. ACTIVE: ImageTrack got enough sensor updates
        3. OCCLUDED: The track is occluded by another object and therefore currently
                     not visible. In this state, the bounding-box of the last update
                     before the occlusion will be used, until it is not occluded / gets
                     a sensor update.
        4. DEAD: If a ImageTrack does not get sensor updates for the configured period
                 of frames
    """

    INITIATED = "INITIATED"
    ACTIVE = "ACTIVE"
    OCCLUDED = "OCCLUDED"
    DEAD = "DEAD"


@dataclass
class TrackEvent:
    """
    Class for storing all tracking information related to a single bounding box at the time of a specific frame.
    """

    bounding_box: BoundingBox
    timestamp: datetime
    state: TrackerState
    frame_id: int
    track_id: int
    speed: float

    __timestamp_format: str = "%Y-%m-%d_%H-%M-%S"

    def __repr__(self) -> str:
        return (
            f"TrackEvent - "
            f"timestamp: {self.timestamp.strftime(self.__timestamp_format)} "
            f"state: {self.state} "
            f"frame_id: {self.frame_id} "
            f"track_id: {self.track_id} "
            f"speed: {self.speed} "
            f"bounding_box: {self.bounding_box}"
        )

    def to_dict(self, raw_type: bool = False, reduced: bool = False) -> Dict[str, Any]:
        """
        Args:
            raw_type: Whether to return the class identifier and timestamp as object or in its representation
                      as dictionary
            reduced: Whether to return the full or a reduced representation of each bounding box

        Returns:
            A dictionary representations of the track event
        """
        return {
            "bounding_box": self.bounding_box.to_dict(
                raw_type=raw_type, reduced=reduced
            ),
            "timestamp": self.timestamp
            if raw_type
            else self.timestamp.strftime(self.__timestamp_format),
            "state": self.state.value,
            "frame_id": self.frame_id,
            "track_id": self.track_id,
            "speed": self.speed,
        }

    @staticmethod
    def from_dict(input_dict: Dict[str, Any], reduced: bool = False) -> TrackEvent:
        """
        Creates a new TrackEvent object from the dictionary representation.

        Args:
            input_dict: The dictionary to create the TrackEvent from
            reduced: Whether the input_dict stores a reduced version of information

        Returns:
            The TrackEvent created from the input_dict
        """
        return TrackEvent(
            bounding_box=BoundingBox.from_dict(
                input_dict=input_dict["bounding_box"], reduced=reduced
            ),
            timestamp=datetime.strptime(
                input_dict["timestamp"], TrackEvent.__timestamp_format
            ),
            state=TrackerState(input_dict["state"]),
            frame_id=int(input_dict["frame_id"]),
            track_id=input_dict["track_id"],
            speed=input_dict["speed"],
        )


class ImageTrack:
    """
    Class using Kalman filter to track a single object represented by its bounding box, storing its
    history as a list of TrackEvents and managing the lifecycle of the track.
    """

    def __init__(
        self,
        configuration: TrackerConfig,
        track_id: int,
        initial_frame_id: int,
        initial_bbox: BoundingBox,
        initial_color_hist: Optional[ImageType] = None,
        meta_info: Optional[Dict[str, Any]] = None,
    ) -> None:
        """
        Initialize object.

        Args:
            initial_frame_id: The time stamp to start with.
            initial_bbox: First detection.
            configuration: Tracker configuration for this ImageTrack object
            initial_color_hist: Color history of the initial box. Can be None if not used.
            meta_info: Dictionary providing meta information for this ImageTrack object
        """

        # ========================================================
        # Static information

        self.configuration = configuration

        self.track_id = track_id

        # The tracks class-identifier is defined by its initial bounding-box
        self.class_identifier: ClassIdentifier = initial_bbox.class_identifier

        # TODO: Store meta_info per TrackEvent?
        # Dictionary that can be filled with meta information about this ImageTrack
        self.meta_info: Optional[Dict[str, Any]] = meta_info

        self._start_frame_id: int = initial_frame_id

        self.current_state: TrackerState = TrackerState.INITIATED

        # Flag that indicates that the ImageTrack has been active for at least one frame
        self.was_active: bool = False

        if self.configuration.min_detections_active == 0:
            self.current_state = TrackerState.ACTIVE
            self.was_active = True

        # ========================================================
        # Dynamic information that changes with the tracking

        # The current color histogram of this ImageTrack
        self.current_color_hist: Optional[ImageType] = initial_color_hist

        # Will be counted up for every call of next_frame(...)
        self.current_frame_id: int = initial_frame_id

        # Current speed for this ImageTrack, defined as the distance of pixels
        # that the track has traveled between two consecutive frames
        self._current_speed: float = 0.0

        # Current speed for this ImageTrack, defined as the distance of pixels
        # that the track has traveled between two consecutive frames
        self._current_bounding_box: BoundingBox = initial_bbox

        # The complete track. Note that the key is the frame ID and can thus
        # start and end at any time but should not have gaps if next_frame is
        # always called.
        # Key: Frame ID
        self.track_events: Dict[int, TrackEvent] = {
            self.current_frame_id: TrackEvent(
                bounding_box=initial_bbox,
                timestamp=datetime.now(),
                state=self.current_state,
                frame_id=self.current_frame_id,
                track_id=self.track_id,
                speed=0.0,
            )
        }

        # The complete information about sensor updates.
        # Initialization counts as one sensor update.
        # Key: Frame ID
        self._sensor_updates: Dict[int, BoundingBox] = {
            self.current_frame_id: deepcopy(initial_bbox)
        }

        self._kf: KalmanFilter = ImageTrack.create_kalman_filter(
            kalman_filter_config=self.configuration.kalman_filter_config,
            initial_bbox=initial_bbox,
        )

    def __repr__(self) -> str:
        return (
            f"ImageTrack: class-identifier={self.class_identifier}, "
            f"len: {len(self.track_events)}, "
            f"latest_track_event: {self.get_latest_track_event()}"
        )

    def to_dict(self, raw_type: bool = False, reduced: bool = False) -> Dict[str, Any]:
        """
        TODO
        Args:
            raw_type:
            reduced:

        Returns:

        """
        return self.track_events[self.current_frame_id].to_dict(
            raw_type=raw_type, reduced=reduced
        )

    def to_json(self) -> Any:
        return self.to_dict(raw_type=False)

    @property
    def current_speed(self) -> float:
        """
        TODO
        Returns:

        """
        return self._current_speed

    @staticmethod
    def create_kalman_filter(
        kalman_filter_config: KalmanFilterConfig, initial_bbox: BoundingBox
    ) -> KalmanFilter:
        """
        Create a KalmanFilter object from the given configuration
        and initial bounding box.

        State is position and speed for x and y with unit pixel and
        pixel per frame so this is independent of fps. The (x, y)
        position is the center of the box that will be tracked with
        the kalman filter.

        - State transition:
         x      1 1 0 0       x
        dx  =   0 1 0 0      dx
         y      0 0 1 1       y
        dy      0 0 0 1      dy

        - Measurement:
         x      1 0
        dx  =   0 0     z_x
         y      0 1     z_y
        dy      0 0

        Args:
            kalman_filter_config: The KalmanFilterConfig defining the relevant parameter
            initial_bbox: The bounding-box for initializing the KalmanFilter

        Returns:
            The created KalmanFilter object
        """

        kalman_filter: KalmanFilter = KalmanFilter(dim_x=4, dim_z=2)

        kalman_filter.x = np.array(
            [[initial_bbox.box.center()[0]], [0], [initial_bbox.box.center()[1]], [0]]
        )

        kalman_filter.H = np.array([[1, 0, 0, 0], [0, 0, 1, 0]])

        kalman_filter.F = np.array(
            [[1, 1, 0, 0], [0, 1, 0, 0], [0, 0, 1, 1], [0, 0, 0, 1]]
        )

        kalman_filter.R = np.array(
            [
                [kalman_filter_config.R, 0],
                [0, kalman_filter_config.R],
            ]
        )

        kalman_filter.P *= kalman_filter_config.P

        # self.kf.Q = Q_discrete_white_noise(dim=4, dt=1, var=0.1)

        return kalman_filter

    def get_start_time(self) -> datetime:
        """
        Returns:
            The datetime object when the track was initially started
        """
        return self.track_events[self._start_frame_id].timestamp

    def get_alive_time(self) -> timedelta:
        """
        Returns:
            The datetime object when the track was initially started
        """
        return datetime.now() - self.track_events[self._start_frame_id].timestamp

    def get_stop_time(
        self,
    ) -> Optional[datetime]:
        """
        Get the stop time of this ImageTrack. When the ImageTrack is still active,
        then it returns None.

        Returns:
            The datetime object when the track was stopped / was not active anymore
        """
        if not self.is_active():
            return self.track_events[max(self.track_events.keys())].timestamp

        return None

    def is_valid(self) -> bool:
        """
        An ImageTrack is valid when it was active for at least one frame

        Returns:
            Whether the ImageTrack is valid
        """
        return self.was_active

    def is_alive(self) -> bool:
        """
        An ImageTrack is alive when it is not in the TrackerState.DEAD state

        Returns:
            Whether a track is alive
        """
        return self.current_state is not TrackerState.DEAD

    def is_active(self) -> bool:
        """
        Determines if a track is still active, based on:

        Returns:
            Whether the track is counted as active based on the described conditions
        """
        return (
            self.current_state is TrackerState.ACTIVE
            or self.current_state is TrackerState.OCCLUDED
        )

    def last_sensor_update_frame_id(self) -> int:
        """
        Determines the last frame ID where a detection was added.
        Can be used to determine how long an object was not detected.

        Returns:
            The frame ID
        """
        return max(self._sensor_updates.keys())

    def set_color_histogram(self, color_hist: Optional[ImageType]) -> None:
        """
        Sets the current color histogram for this track

        Args:
            color_hist: The color histogram to set

        Returns:
            None
        """
        alpha = self.configuration.assignment_cost_config.color_cost.color_filter_alpha
        if self.current_color_hist is not None and alpha < 1:
            self.current_color_hist = (1 - alpha) * self.current_color_hist + alpha * color_hist  # type: ignore
        else:
            self.current_color_hist = color_hist

    def get_latest_track_event(
        self,
    ) -> TrackEvent:
        """
        Determines the latest TrackEvent of this track.

        Returns:
            The determined TrackEvent
        """

        return self.track_events[max(self.track_events.keys())]

    def get_current_bounding_box(self) -> BoundingBox:
        return self._current_bounding_box

    def get_redetect_radius(self) -> float:
        """
        Based on the motion model of objects, this determines a radius where the object could
        have been moved to while it was not detected / got no sensor updates. Note that the
        kalman filter only gives the estimated position based on the last speed. A kalman filter
        does not include such a motion model as applied here.

        Returns:
            The possible radius the object can be found now.
        """

        # TODO discuss readding s0 to relax/widen the radius for redetection
        # Formula:
        # s = v*t + s0  => s0 is always zero, since the distance is measured from the
        #                  center of the bounding.box

        object_speed_config: TrackingToolTrackerConfigObjectSpeed = (
            self.configuration.assignment_cost_config.distance_cost.obj_speed[
                self.class_identifier.class_id
            ]
        )

        return (
            self.current_frame_id - self.last_sensor_update_frame_id()
        ) * object_speed_config.x + object_speed_config.b

    def __update_current_bounding_box(self) -> None:
        """
        Use the bounding-boxes for <kalman_delay> frames as tracker positions
        instead of the position from the filter.
        Reason: We need some detections to be able to estimate a good speed.
        """
        if self.configuration.kalman_filter_config.kalman_delay >= len(
            self._sensor_updates
        ):
            self._current_bounding_box = deepcopy(
                self._sensor_updates[self.last_sensor_update_frame_id()]
            )
        else:
            self._current_bounding_box.box.new_center(
                int(self._kf.x[0]), int(self._kf.x[2])
            )

    def __update_speed_and_track_events(self) -> None:
        # update speed
        if len(self.track_events) >= 2:
            self._current_speed = euclidean_distance(
                box_1=self._current_bounding_box.box,
                box_2=self.track_events[self.current_frame_id - 1].bounding_box.box,
            )

        # update the current TrackEvent
        self.track_events[self.current_frame_id] = TrackEvent(
            bounding_box=self._current_bounding_box,
            timestamp=datetime.now(),
            state=self.current_state,
            frame_id=self.current_frame_id,
            track_id=self.track_id,
            speed=self.current_speed,
        )

    def predict(
        self, occlusion_bounding_boxes: Optional[List[BoundingBox]] = None
    ) -> None:
        """
        Update the internal Kalman filter (prediction step) and the state of this ImageTrack.
        It does not have an effect if the ImageTrack is in the TrackerState.DEAD state.

        Must be called for every frame in the following order:
        predict(...) --> (Optional) update(...)

        Args:
            occlusion_bounding_boxes: Potential bounding boxes of object that might occlude
                                      this track, respectively the current bounding box of
                                      this track

        Returns:
            None
        """

        if self.current_state is TrackerState.DEAD:
            return

        last_sensor_update_frame_id = self.last_sensor_update_frame_id()

        # Check if ImageTrack got to old and therefore is DEAD now
        if (
            self.current_frame_id - last_sensor_update_frame_id
            > self.configuration.max_age
        ):
            self.current_state = TrackerState.DEAD
            self.track_events[max(self.track_events.keys())].state = TrackerState.DEAD

            # return prematurely because the image track is DEAD now
            return

        self.current_frame_id += 1
        self._kf.predict()

        self.__update_current_bounding_box()

        # check occlusion
        if (
            occlusion_bounding_boxes is not None
            and self.current_state is TrackerState.ACTIVE
        ):
            for occlusion_bounding_box in occlusion_bounding_boxes:
                # TODO: Check if this could trigger to much computing time
                if (
                    compute_iou(
                        box_1=occlusion_bounding_box.box,
                        box_2=self._current_bounding_box.box,
                    )
                    > 0.0
                ):
                    # Assume that the bounding box has not changed during the time
                    # the object was occluded
                    self.current_state = TrackerState.OCCLUDED
                    break

        self.__update_speed_and_track_events()

    def update(self, bounding_box: BoundingBox) -> None:
        """
        Perform a sensor update of the internal kalman filter for this
        ImageTrack using the given bounding-box. It does not have an effect
        if the ImageTrack is in the TrackerState.DEAD state

        Should be called once within a step in the following order:
        predict(...) --> (Optional) update(...)

        Args:
            bounding_box: The bounding box that should be used as sensor update

        Returns:
            None
        """

        if self.current_state is TrackerState.DEAD:
            return

        if self.current_state is TrackerState.OCCLUDED:
            self.current_state = TrackerState.ACTIVE

        self._sensor_updates[self.current_frame_id] = deepcopy(bounding_box)
        self._kf.update(bounding_box.box.center())

        # Switch to state ACTIVE when the ImageTrack got enough sensor updates
        if (
            self.current_state is TrackerState.INITIATED
            and len(self._sensor_updates) >= self.configuration.min_detections_active
        ):
            self.current_state = TrackerState.ACTIVE
            self.was_active = True

        self.__update_current_bounding_box()
        self.__update_speed_and_track_events()



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

import logging
from datetime import datetime
from typing import Dict, List

from mlcvzoo_base.api.data.bounding_box import BoundingBox
from mlcvzoo_base.api.data.box import Box
from mlcvzoo_base.api.data.class_identifier import ClassIdentifier
from mlcvzoo_base.configuration.structs import ObjectDetectionBBoxFormats
from mlcvzoo_base.configuration.utils import str2bool
from mlcvzoo_base.utils import ensure_dir

from mlcvzoo_tracker.image_track import TrackerState, TrackEvent

logger = logging.getLogger(__name__)


def write_image_tracks_to_mot(
    complete_pallet_track_events: Dict[int, List[TrackEvent]],
    output_file_path: str,
) -> None:
    """
    Write an annotation file in the MOT201617 / MOT2020 format
    based on the given list of image-tracks. Format for is:

    <frame>, <id>, <bb_left>, <bb_top>, <bb_width>, <bb_height>, <conf>, <class>, <visibility>

    Args:
        complete_pallet_track_events: 1st key: Frame-ID
                                      2nd key: Track-ID
                                      value: TrackEvent
        output_file_path: The path to the file were the data should be written to

    Returns:
        None
    """

    mot_lines: List[str] = []

    for frame_id, track_events in complete_pallet_track_events.items():
        for track_event in track_events:
            if (
                track_event.state is TrackerState.ACTIVE
                or track_event.state is TrackerState.OCCLUDED
                or track_event.state is TrackerState.INITIATED
            ):
                conf = "1.0"
                visibility = "1.0"
                # IMPORTANT NOTE: Indices in MOT are 1-based, therefore plus 1 for the IDs
                mot_lines.append(
                    f"{frame_id + 1},"
                    f"{track_event.track_id + 1},"
                    f"{track_event.bounding_box.box.xmin},"
                    f"{track_event.bounding_box.box.ymin},"
                    f"{track_event.bounding_box.box.width},"
                    f"{track_event.bounding_box.box.height},"
                    f"{conf},"
                    f"{track_event.bounding_box.class_id + 1},"
                    f"{visibility}\n"
                )

    ensure_dir(file_path=output_file_path, verbose=True)

    logger.debug("Write mot annotation file to '%s'" % output_file_path)
    with open(output_file_path, "w") as output_file:
        output_file.writelines(mot_lines)


def read_image_tracks(input_path: str) -> Dict[int, List[TrackEvent]]:
    """
    Read an annotation file in the MOT201617 / MOT2020 format
    Format for MOT is:

    <frame>, <id>, <bb_left>, <bb_top>, <bb_width>, <bb_height>, <conf>, <class>, <visibility>

    Args:
        input_path:

    Returns:

    """
    complete_pallet_track_events: Dict[int, List[TrackEvent]] = {}

    with open(input_path, "r") as input_file:
        mot_lines = input_file.readlines()

    for mot_line in mot_lines:
        mot_line_split = mot_line.strip().split(",")

        # IMPORTANT NOTE: Indices in MOT are 1-based, therefore minus 1 for the IDs
        frame_id = int(mot_line_split[0]) - 1

        if frame_id not in complete_pallet_track_events:
            complete_pallet_track_events[frame_id] = []

        complete_pallet_track_events[frame_id].append(
            TrackEvent(
                bounding_box=BoundingBox(
                    box=Box.init_format_based(
                        box_list=(
                            int(float(mot_line_split[2])),
                            int(float(mot_line_split[3])),
                            int(float(mot_line_split[4])),
                            int(float(mot_line_split[5])),
                        ),
                        box_format=ObjectDetectionBBoxFormats.XYWH,
                    ),
                    class_identifier=ClassIdentifier(
                        class_id=int(mot_line_split[7]) - 1, class_name="MOT"
                    ),
                    score=float(mot_line_split[6]),
                    difficult=False,
                    occluded=str2bool(mot_line_split[8]),
                    content="",
                ),
                timestamp=datetime.now(),
                state=TrackerState.ACTIVE,
                frame_id=frame_id,
                track_id=int(mot_line_split[1]) - 1,
                speed=0.0,
            )
        )

    return complete_pallet_track_events



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""
Module that defines an multi object tracker based on bounding boxes. It utilizes
the ImageTrack class and determines an optimal bounding box to track assignment
by using the principle of the "hungarian algorithm" for minimizing a cost matrix.
"""

import logging
import time
from typing import Any, Dict, List, Optional

import cv2
import numpy as np
from mlcvzoo_base.api.data.bounding_box import BoundingBox
from mlcvzoo_base.api.data.box import compute_iou, euclidean_distance
from mlcvzoo_base.api.data.class_identifier import ClassIdentifier
from nptyping import Float, NDArray, Shape
from scipy.optimize import linear_sum_assignment

from mlcvzoo_tracker.configuration import TrackerConfig
from mlcvzoo_tracker.image_track import ImageTrack
from mlcvzoo_tracker.types import ImageType

logger = logging.getLogger(__name__)


class HungarianImageTracker:
    """
    Class for matching newly detected bounding boxes to existing tracks via the hungarian algorithm.
    """

    def __init__(
        self, configuration: TrackerConfig, object_class_identifier: ClassIdentifier
    ):
        """
        Initialize this object.
        """

        self.configuration = configuration
        self.current_frame_id: int = 0
        self._tracks: List[ImageTrack] = []
        self.object_class_identifier = object_class_identifier

        self.track_id_counter: int = 0

    def __str__(self) -> str:
        return "{} tracks".format(len(self._tracks))

    def to_list(self, raw_type: bool = False, reduced: bool = False) -> List[Any]:
        """
        Args:
            raw_type: Whether to return the class identifier and timestamp as object or in its representation
                      as dictionary
            reduced: Whether to return the full or a reduced representation of each bounding box and timestamp

        Returns:
            A list of dictionary representations of all stored tracks.
        """
        return [t.to_dict(raw_type=raw_type, reduced=reduced) for t in self._tracks]

    def compute_track_statistics(self) -> Dict[int, Dict[str, Dict[str, Any]]]:
        track_statistics: Dict[int, Dict[str, Dict[str, Any]]] = {}
        for track in self.get_tracks():
            speed_list: List[float] = []
            tracks_widths: List[int] = []
            tracks_heights: List[int] = []
            for track_event in track.track_events.values():
                if track_event.speed > 0.0:
                    speed_list.append(track_event.speed)
                tracks_widths.append(track_event.bounding_box.box.width)
                tracks_heights.append(track_event.bounding_box.box.height)

            if not (
                len(speed_list) > 0
                and len(tracks_widths) > 0
                and len(tracks_heights) > 0
            ):
                continue

            track_statistics[track.track_id] = {
                "speed_info": {
                    "max_speed": max(speed_list),
                    "min_speed": min(speed_list),
                    "avg_speed": sum(speed_list) / len(speed_list),
                },
                "width_info": {
                    "max_width": max(tracks_widths),
                    "min_width": min(tracks_widths),
                    "avg_width": sum(tracks_widths) / len(tracks_widths),
                },
                "height_info": {
                    "max_height": max(tracks_heights),
                    "min_height": min(tracks_heights),
                    "avg_height": sum(tracks_heights) / len(tracks_heights),
                },
            }

        return track_statistics

    def get_valid_tracks(self) -> List[ImageTrack]:
        """
        Get tracks that had enough bounding_boxes updates to be valid and are currently
        still alive (got bounding_boxes lastly).

        Returns:
            All tracks that are valid
        """

        return [track for track in self._tracks if track.is_valid()]

    def get_active_tracks(self) -> List[ImageTrack]:
        """
        Returns:
            All tracks that are active
        """
        return [t for t in self._tracks if t.is_active()]

    def get_alive_tracks(self) -> List[ImageTrack]:
        """
        Returns:
            All tracks that are alive
        """
        return [track for track in self._tracks if track.is_alive()]

    def get_tracks(self) -> List[ImageTrack]:
        """
        Returns:
            All currently managed ImageTrack's of the tracker.
        """
        return self._tracks

    @staticmethod
    def _compute_iou_costs(
        iou_weight: float,
        prev_bbox: Optional[BoundingBox],
        cur_bbox: Optional[BoundingBox],
    ) -> float:
        iou_cost: float = 0.0

        if iou_weight > 0 and prev_bbox and cur_bbox:
            iou_cost = 1.0 - compute_iou(box_1=prev_bbox.box, box_2=cur_bbox.box)

        return iou_cost

    @staticmethod
    def _compute_norm_dist_cost(
        distance_weight: float,
        max_dist: float,
        prev_bbox: Optional[BoundingBox],
        cur_bbox: Optional[BoundingBox],
    ) -> float:
        norm_dist_cost = 0.0
        if distance_weight > 0 and prev_bbox and cur_bbox:
            norm_dist_cost = euclidean_distance(prev_bbox.box, cur_bbox.box) / max_dist

        return norm_dist_cost

    @staticmethod
    def _compute_color_cost(
        color_weight: float,
        prev_hist: Optional[ImageType],
        cur_hist: Optional[ImageType],
    ) -> float:
        # TODO should this be a config parameter, is 10 a reasonable value?
        color_dist_cost = 10.0  # In case it cannot be calculated

        if color_weight > 0 and prev_hist is not None and cur_hist is not None:
            color_dist_cost = cv2.compareHist(prev_hist, cur_hist, 4)

        return color_dist_cost

    @staticmethod
    def _compute_cost_matrix(
        iou_weight: float,
        distance_cost_weight: float,
        color_cost_weight: float,
        alive_tracks: List[ImageTrack],
        bounding_boxes: List[BoundingBox],
        current_hists: List[Optional[ImageType]],
    ) -> ImageType:
        # Span cost Matrix
        # row => alive_tracks
        # column => new bounding boxes

        cost_matrix = np.zeros((len(alive_tracks), len(bounding_boxes)))
        for row, alive_track in enumerate(alive_tracks):
            # TODO: Get the ImageTrack bounding boxes as parameter?
            prev_bbox = alive_track.get_current_bounding_box()
            for col, (cur_bbox, cur_hist) in enumerate(
                zip(bounding_boxes, current_hists)
            ):
                # IoU costs
                iou_cost: float = HungarianImageTracker._compute_iou_costs(
                    iou_weight=iou_weight,
                    prev_bbox=prev_bbox,
                    cur_bbox=cur_bbox,
                )

                # TODO no consequences when exceeding max_dist?

                # Distance costs
                norm_dist_cost: float = HungarianImageTracker._compute_norm_dist_cost(
                    distance_weight=distance_cost_weight,
                    max_dist=alive_track.get_redetect_radius(),
                    prev_bbox=prev_bbox,
                    cur_bbox=cur_bbox,
                )

                # Color costs
                color_dist_cost: float = HungarianImageTracker._compute_color_cost(
                    color_weight=color_cost_weight,
                    prev_hist=alive_track.current_color_hist,
                    cur_hist=cur_hist,
                )

                # Complete weighted cost
                cost_matrix[row, col] = (
                    iou_weight * iou_cost
                    + distance_cost_weight * norm_dist_cost
                    + color_cost_weight * color_dist_cost
                )

        return cost_matrix

    @staticmethod
    def _update_tracks(
        row_ind: NDArray[Shape["Any"], Float],
        col_ind: NDArray[Shape["Any"], Float],
        assignment_threshold: float,
        cost_matrix: NDArray[Shape["Any, Any"], Float],
        alive_tracks: List[ImageTrack],
        bounding_boxes: List[BoundingBox],
        current_hists: List[Optional[ImageType]],
    ) -> List[bool]:
        """
        Calls the update(...) method for all object in the alive-track list.


        Args:
            row_ind:
            col_ind:
            assignment_threshold:
            cost_matrix:
            alive_tracks:
            bounding_boxes:
            current_hists:

        Returns:

        """
        #
        # t1 : b1 - - - bm
        #  |
        #  |
        # tn : b1 - - - bm

        # List that states whether a bounding box entry has been assigned to a track
        are_detections_assigned: List[bool] = len(bounding_boxes) * [False]

        for i in range(len(row_ind)):
            bbox_index = col_ind[i]
            track_index = row_ind[i]

            if (
                cost_matrix[track_index, bbox_index] < assignment_threshold
                and not are_detections_assigned[bbox_index]
            ):
                _t = alive_tracks[row_ind[i]]
                _t.update(bounding_boxes[bbox_index])
                if current_hists[bbox_index] is not None:
                    _t.set_color_histogram(current_hists[bbox_index])
                are_detections_assigned[bbox_index] = True

        return are_detections_assigned

    def next(
        self,
        bounding_boxes: List[BoundingBox],
        frame: Optional[ImageType] = None,
        occlusion_bounding_boxes: Optional[List[BoundingBox]] = None,
    ) -> None:
        """
        Update the internal list of ImageTracks by using
        the given bounding-box and frame information

        Args:
            bounding_boxes: All bounding-boxes for this frame that should be considered
                            as sensor update
            frame: (Optional) The complete image for cutting out
                   color histograms from the given bounding boxes
            occlusion_bounding_boxes: bounding-boxes that can possibly be a reason for
                                      occluding an ImageTrack for this class

        Returns:
            None
        """

        log_timings: bool = False

        current_hists: List[Optional[ImageType]]

        start = time.time()
        if (
            frame is not None
            and self.configuration.assignment_cost_config.color_cost.weight > 0.0
        ):
            # TODO: Check for runtime issues
            current_hists = [
                b.box.color_hist(
                    margin_x=self.configuration.assignment_cost_config.color_cost.margin_x,
                    margin_y=self.configuration.assignment_cost_config.color_cost.margin_y,
                    frame=frame,
                )
                for b in bounding_boxes
            ]
        else:
            current_hists = len(bounding_boxes) * [None]  # type: ignore[assignment]

        if log_timings:
            logger.info("color_hist: %s" % (time.time() - start))

        # List that states whether a bounding box entry has been assigned to a track
        are_detections_assigned: List[bool]

        # Do this if there are already tracks, not in case of first frame
        if self._tracks:
            alive_tracks = self.get_alive_tracks()

            for track in alive_tracks:
                track.predict(occlusion_bounding_boxes=occlusion_bounding_boxes)

            start = time.time()
            cost_matrix: NDArray[
                Shape["Any, Any"], Float
            ] = HungarianImageTracker._compute_cost_matrix(
                iou_weight=self.configuration.assignment_cost_config.iou_weight,
                distance_cost_weight=self.configuration.assignment_cost_config.distance_cost.weight,
                color_cost_weight=self.configuration.assignment_cost_config.color_cost.weight,
                alive_tracks=alive_tracks,
                bounding_boxes=bounding_boxes,
                current_hists=current_hists,
            )
            if log_timings:
                logger.info("_compute_cost_matrix: %s" % (time.time() - start))

            # Based on the cost-matrix run a linear-sum-assignment that is based on the
            # Hungarian algorithm. This solves the problem of assigning new bounding-boxes
            # to active tracks.
            row_ind: NDArray[Shape["Any"], Float]
            col_ind: NDArray[Shape["Any"], Float]
            start = time.time()
            row_ind, col_ind = linear_sum_assignment(cost_matrix)
            if log_timings:
                logger.info("linear_sum_assignment: %s" % (time.time() - start))

            start = time.time()
            are_detections_assigned = HungarianImageTracker._update_tracks(
                row_ind=row_ind,
                col_ind=col_ind,
                assignment_threshold=self.configuration.assignment_cost_config.assignment_threshold,
                cost_matrix=cost_matrix,
                alive_tracks=alive_tracks,
                bounding_boxes=bounding_boxes,
                current_hists=current_hists,
            )
            if log_timings:
                logger.info("_update_tracks: %s" % (time.time() - start))
        else:
            are_detections_assigned = len(bounding_boxes) * [False]

        # In case the cost matrix was not square there are unassigned
        # detections or tracks. Therefore:
        # - Create a new track for each unassigned detection
        # - Not updated tracks are just ignored
        for det_idx, (is_assigned, bounding_box) in enumerate(
            zip(are_detections_assigned, bounding_boxes)
        ):
            if not is_assigned:
                self._tracks.append(
                    ImageTrack(
                        configuration=self.configuration,
                        track_id=self.track_id_counter,
                        initial_frame_id=self.current_frame_id,
                        initial_bbox=bounding_box,
                        initial_color_hist=current_hists[det_idx],
                    )
                )
                self.track_id_counter += 1

        self.current_frame_id += 1

        if not self.configuration.keep_dead_tracks:
            self._tracks = [t for t in self._tracks if t.is_alive()]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mlcvzoo_tracker']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=20',
 'filterpy>=1.4',
 'mlcvzoo_base>=5.1,<6.0',
 'numpy>=1.19.2,!=1.19.5',
 'opencv-contrib-python>=4.5,!=4.5.5.64',
 'opencv-python>=4.5,!=4.5.5.64',
 'related-mltoolbox>=1.0,<2.0',
 'yaml-config-builder>=8']

setup_kwargs = {
    'name': 'mlcvzoo-tracker',
    'version': '0.1.1',
    'description': 'MLCVZoo Tracker Package',
    'long_description': '# MLCVZoo Tracker\n\nThe MLCVZoo is an SDK for simplifying the usage of various (machine learning driven)\ncomputer vision algorithms. The package **mlcvzoo_tracker** contains tools for tracking detected objects.\n\nFurther information about the MLCVZoo can be found [here](../README.md).\n\n## Install\n`\npip install mlcvzoo-tracker\n`\n\n## Technology stack\n\n- Python\n',
    'author': 'Maximilian Otten',
    'author_email': 'maximilian.otten@iml.fraunhofer.de',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://git.openlogisticsfoundation.org/silicon-economy/base/ml-toolbox/mlcvzoo',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""
Definition of the TrackerConfig
"""

import logging
from typing import List

import related
from attr import define
from config_builder import BaseConfigClass

logger = logging.getLogger(__name__)


@define
class TrackingToolTrackerConfigObjectSpeed(BaseConfigClass):
    class_id: int = related.IntegerField(required=False)
    x: int = related.IntegerField(required=False)
    b: int = related.IntegerField(required=False, default=0)


@define
class KalmanFilterConfig(BaseConfigClass):
    # Kalman filter related parameter

    # Parameter for the kalman filter itself
    R: int = related.IntegerField(required=False, default=1000)
    P: int = related.IntegerField(required=False, default=10)

    # Use the bounding-boxes for <kalman_delay> frames as tracker positions
    # instead of the position from the filter.
    # Reason: We need some detections to be able to estimate a good speed.
    kalman_delay: int = related.IntegerField(required=False, default=10)


@define
class DistanceCostConfig(BaseConfigClass):
    weight: float = related.FloatField(required=False, default=1.0)

    # radius for assignment of detection to track = time not seen * x + b
    obj_speed: List[TrackingToolTrackerConfigObjectSpeed] = related.SequenceField(
        TrackingToolTrackerConfigObjectSpeed, required=False, default=[]
    )


@define
class ColorCostConfig(BaseConfigClass):
    # Color histogram related parameter
    weight: float = related.FloatField(required=False, default=0.0)

    # Margin for cropping the image
    margin_x: float = related.FloatField(required=False, default=0.5)
    margin_y: float = related.FloatField(required=False, default=0.5)
    # Alpha value for the histogram
    color_filter_alpha: float = related.FloatField(required=False, default=0)


@define
class AssignmentCostConfig(BaseConfigClass):
    color_cost: ColorCostConfig = related.ChildField(
        cls=ColorCostConfig, required=False, default=ColorCostConfig()
    )

    distance_cost: DistanceCostConfig = related.ChildField(
        cls=DistanceCostConfig, required=False, default=DistanceCostConfig()
    )

    # Threshold for the total weights that determining if a
    # bounding-box should be assigned to a track, sum of:
    # - IoU costs,
    # - Color Histogram costs
    # - Distance costs
    assignment_threshold: float = related.FloatField(required=False, default=1.5)

    iou_weight: float = related.FloatField(required=False, default=1.0)


@define
class TrackerConfig(BaseConfigClass):
    kalman_filter_config: KalmanFilterConfig = related.ChildField(
        cls=KalmanFilterConfig, required=False, default=KalmanFilterConfig()
    )

    assignment_cost_config: AssignmentCostConfig = related.ChildField(
        cls=AssignmentCostConfig, required=False, default=AssignmentCostConfig()
    )

    # Minimum amount of bounding-box updates for a track to be ACTIVE
    min_detections_active: int = related.IntegerField(required=False, default=5)

    # The maximum number of frames a track is allowed to get no sensor updates.
    # If this value is exceeded, the track is counted as DEAD
    max_age: int = related.IntegerField(required=False, default=20)

    # Whether to keep ImageTracks with the state DEAD
    keep_dead_tracks: bool = related.BooleanField(required=False, default=True)

    def check_values(self) -> bool:
        if self.min_detections_active < 0:
            logger.error("The minimum value for min_detections_active is zero.")
            return False

        if self.max_age < 0:
            logger.error(
                "The minimum value for max_age is zero. In this case, a sensor update is required in each iteration to keep the trace alive."
            )
            return False

        return True



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

__version__ = "0.1.1"
__license__ = "OLFL-1.3"
